/* 
 * File:   bbclass.cpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 * 
 * Created on April 3, 2014, 8:26 PM
 */

#include "bbclass.hpp"
#include "nodeclass.hpp"
#include <cstdio>

using namespace std;

bbclass::bbclass() {
}

bbclass::bbclass(const bbclass& orig) {
}

bbclass::~bbclass() {
}

int bbclass::optimize() {

    numProc=communicator.Get_size();
    initial_time = MPI::Wtime();
    
    cout<<"Algorithm starts at "<<initial_time<<endl;

    int nodes_sent;
    int sendProc;
    int count_received;
    
    double sendTime;
    double receiveTime;
    
    nodestruct *current = new nodestruct[1];

    solution_status = new int[numProc];
    solution_value = new double[numProc];
    branch_variable = new int[numProc];

    nodestruct **node_to_processor = new nodestruct*[numProc];
    
    
    int it_counter = 0;
    
    while (nodes_open > 0 and (MPI::Wtime()-initial_time)<36000) {
        
        /*int shownodes = show_nodes();
        if (shownodes<nodes_open) {
            cout<<"number of open problems"<<nodes_open<<"while list size is"<<shownodes<<endl;
            return 0;
        }*/
        
        nodes_sent = 0;
        
        current = first_node;
        
        // Counter for sending problems to workers
        sendProc=numProc-1;
        
        // Create array to save which processor receives which nodestruct
        
        while (sendProc>0 and nodes_open>0) {
            
            sendTime = MPI::Wtime();
            // If a node turns out to have a lower bound higher than the incumbent value, prune this            
            if (incumbent_node != NULL && current->node.get_lowerbound() > incumbent_node->node.get_value()) {
                nodestruct *node_to_prune = current;
                current = current->next;
                prune_node(node_to_prune);
                nodes_open--;
            }
            else {

                // Save which processor receives which nodestruct
                node_to_processor[sendProc] = current;

                //current->node.set_processor(sendProc);

               // Send notification
                send_notification(1, sendProc);

                int *fixed_vars = new int[num_bins];
                current->node.get_binaries(fixed_vars, num_bins);

                //send the current nodestruct to a processor
                send_binaries(fixed_vars, sendProc);

                delete [] fixed_vars;

                //update processor_count
                sendProc--;
                nodes_open--;
                nodes_sent++;
            
                // Set current node to the next node in the list
                if (current->next != NULL) {

                    //cout<<" update current from "<<current->node.get_tag()<<" to "<<current->next->node.get_tag()<<endl;

                    current = current->next;
                }

                else {
                    break;
                }
            
               // cout<<myRank<<": next available processor : "<<sendProc<<" open problems :"<<nodes_open<<endl;
            }
            
            //cout<<"Sending time was "<<MPI::Wtime()-sendTime<<" seconds."<<endl;


        }
        
        //receive solutions from workers
        count_received = 0;
        
        while (count_received < nodes_sent) {
            // Wait until something is received
            while (communicator.Iprobe(MPI::ANY_SOURCE, MPI::ANY_TAG, status) == 0);
            if (count_received == 0) {
            receiveTime = MPI::Wtime();
            }
            // Save identity of origin worker
            int origin_worker = status.Get_source();
                                    
            // Receive the solution status and value if any
            receive_solution(origin_worker);
             //cout<<myRank<<": received from: "<<origin_worker<<" solution status: "<<solution_status[origin_worker]<<endl;

            //Prune or branch according to solution_status received
            if (solution_status[origin_worker]==0) {
                prune_node(node_to_processor[origin_worker]);
            }
              
            else if (solution_status[origin_worker] == 2  ) {
                cout<<" integer solution found with value "<<solution_value[origin_worker]<<" and fixed variables "<<endl;
                int *fixed = new int[num_bins];
                node_to_processor[origin_worker]->node.get_binaries(fixed, num_bins);
                for (int i=0;i<num_bins;i++) {
                    cout<<fixed[i]<<endl;
                }
                if (incumbent_node == NULL) {
                    incumbent_node = node_to_processor[origin_worker];
                    incumbent_node->node.set_value(solution_value[origin_worker]);
                    prune_node(node_to_processor[origin_worker]);
                    cout<<"First incumbent node is "<<incumbent_node->node.get_tag()<<" with value "<<incumbent_node->node.get_value()<<" received from worker "<<origin_worker<<endl;

                }
                else if (solution_value[origin_worker]<incumbent_node->node.get_value() ) {
                    incumbent_node = node_to_processor[origin_worker];
                    incumbent_node->node.set_value(solution_value[origin_worker]);
                    prune_node(node_to_processor[origin_worker]);
                    cout<<"New incumbent node is "<<incumbent_node->node.get_tag()<<" with value "<<incumbent_node->node.get_value()<<" received from worker "<<origin_worker<<endl;
                }
                else {
                    prune_node(node_to_processor[origin_worker]);
                    cout<<"Integer node "<<node_to_processor[origin_worker]->node.get_tag()<<" with value "<<node_to_processor[origin_worker]->node.get_value()<<" was pruned" <<endl;
                }
            }
            
            else if (incumbent_node != NULL and solution_value[origin_worker]>=incumbent_node->node.get_value()) {
                    prune_node(node_to_processor[origin_worker]);
                    cout<<" Node "<<node_to_processor[origin_worker]->node.get_tag()<<" was pruned because value was higher than incumbent value"<<endl;
            }
              
            else if(solution_status[origin_worker] == 3 ) {
                cout<<"Something went wrong in xpress"<<endl;
                return 0;
            }
                
            else {
                // Set the node value to be the solution value received
                node_to_processor[origin_worker]->node.set_value(solution_value[origin_worker]);
                branch_node(node_to_processor[origin_worker], origin_worker);
            }
            
           // cout<<myRank<<":  done branching, open nodes: "<<nodes_open<<endl;
                        
        count_received++;
                   // cout<<myRank<<":  I have received: "<<count_received<<" and I sent "<<nodes_sent<<endl;

        
        }
        
        // cout<<myRank<<":  done branching, open nodes: "<<nodes_open<<endl;


        // Set lower bound equal to the first node's lower bound
        current = first_node;
        
        if (incumbent_node != NULL) {
            lower_bound = incumbent_node->node.get_value();
        }
        
        else {
        lower_bound = current->node.get_lowerbound();
        }
        
        // Check all open nodes if they have a lower bound below this
        for(int i=0;i<nodes_open;i++) {
            if (current == NULL) {
              //  cout<<myRank<<": There is no next node!"<<endl;
                break;
            }
            
            else{
           
                if (current->node.get_lowerbound() < lower_bound) {
                    lower_bound = current->node.get_lowerbound();
                }
                
                current=current->next;              

            }
        }

        // If current lower_bound is equal to or larger than the incumbent value, we have found the optimal solution
        if (incumbent_node != NULL and lower_bound >= incumbent_node->node.get_value()-0.01) {
            break;
        }
        
        it_counter++; 
        cout<<"iteration"<<it_counter<<" lower bound is "<<lower_bound<<endl;
        if (incumbent_node != NULL) {
        cout<<" incumbent value is: "<<incumbent_node->node.get_value()<<endl;
        }

        
    }
    
        delete [] node_to_processor;
        delete [] solution_status;
        delete [] solution_value;
        delete [] branch_variable;
        

    
    cout<<" lower bound is: "<<lower_bound++<<endl;
    if (incumbent_node != NULL) {
        cout<<" incumbent value is: "<<incumbent_node->node.get_value()<<endl;
        cout<<" incumbent node binaries are: "<<endl;
        int *write_binaries = new int(num_bins);
        incumbent_node->node.get_binaries(write_binaries, num_bins);

        for(int i=0;i<num_bins;i++) {
            cout<<" binary number "<<i<<" fixed to "<<write_binaries[i]<<endl;
        }
    }
        
    //Once the list is empty, select the optimal solution and report 0 to workers (terminate)
    for(int i=1;i<numProc;i++) {
        send_notification(0,i);
    }
    
   // cout<<"Algorithm has run for "<<MPI::Wtime()-initial_time<<endl;

        
    return 0;
}

void bbclass::initialize(char *argv, MPI::Intracomm comm){
    
    communicator = comm;
    
    myRank = communicator.Get_rank();
    
    //declare an xpress problem object
    XPRSprob prob;
    //allocate memory using the dedicated function
    XPRScreateprob(&prob);
    //read the problem from the file path provided since main
    XPRSreadprob(prob, argv, "");
    //read the number of binary variables in the problem
    int nCols = 0;
    XPRSgetintattrib(prob, XPRS_COLS, &nCols);
    
    char *types = new char[nCols];
    num_bins=0;
    XPRSgetcoltype(prob, types, 0, nCols-1);
    for(int i=0;i<nCols;i++) if (types[i]=='B') num_bins++;
    
    //start the node list by pointing the first_node pointer to a new node with 
    //all it's variables relaxed.
    int *fix_relax = new int[num_bins];   //allocate memory
    for(int i=0;i<num_bins;i++) fix_relax[i]=-1;   //fill the "all relaxed" array
    
    nodestruct *list_node = new nodestruct[1];                 //create the first nodestruct
    list_node->node.initialize(fix_relax, num_bins, 0);       //call the initialization method for the node member of the struct
    list_node->next = NULL;              //for safety, point the next member of the struct to where it can't hurt
    list_node->previous = NULL;              //for safety, point the next member of the struct to where it can't hurt

    
    first_node = list_node;            //make the "first_node" field of the class point to this new node
    nodes_open=1;               //we now have 1 node!
    incumbent_node = NULL;      //this new node hasn't been solved, so we can't know if it's a solution or not. for problems with unimodular matrix, the first node will always be a solution, and the optimal one at all!
    
    delete [] types;
    delete [] fix_relax;
    XPRSdestroyprob(prob);
}

// Function to send notification to worker, 0 means terminate, 1 means solve
void bbclass::send_notification(int notification, int proc) {
    request = communicator.Issend(&notification, 1, MPI::INT, proc, 0);
    request.Wait(status);
}

// Function to send the status of binary variables to a worker
void bbclass::send_binaries(int* binaries, int proc) {
    request = communicator.Issend(binaries, num_bins , MPI::INT, proc, 1);
    request.Wait(status);
}

// Function to receive solution_status from a worker, 0 means infeasible solution, 1 means LP feasible solution and 2 means integer feasible solution
void bbclass::receive_solution(int origin_worker) {

    communicator.Recv(&solution_status[origin_worker], 1, MPI::INT, origin_worker, 2, status);
             
    // If the solution is feasible, the value is received as well
    if (solution_status[origin_worker]==1) {
        communicator.Recv(&solution_value[origin_worker], 1, MPI::DOUBLE, origin_worker, 3, status);
                      //cout<<myRank<<": received from: "<<origin_worker<<" solution value: "<<solution_value[origin_worker]<<endl;

        communicator.Recv(&branch_variable[origin_worker], 1, MPI::INT, origin_worker, 4, status);

    }
    
    else if (solution_status[origin_worker]==2) {
        communicator.Recv(&solution_value[origin_worker], 1, MPI::DOUBLE, origin_worker, 3, status);
                      //cout<<myRank<<": received from: "<<origin_worker<<" solution value: "<<solution_value[origin_worker]<<endl;

    }
    
    else {
    //some error handling needed!
    }
}

void bbclass::branch_node(nodestruct *node_to_branch, int origin_worker) {
    
    double parent_value = node_to_branch->node.get_value();

    int parent_tag = node_to_branch->node.get_tag();
    
    // Create nodes for children
    //nodeclass *child_down = new nodeclass[1];
    //nodeclass *child_up = new nodeclass[1];
    nodeclass child_down, child_up;

    // Create binary arrays for both children
    int *binaries_down = new int[num_bins];
    int *binaries_up = new int[num_bins];
    
    node_to_branch->node.get_binaries(binaries_down, num_bins);
    node_to_branch->node.get_binaries(binaries_up, num_bins);

    // Choose a variable to branch on
    binaries_down[branch_variable[origin_worker]]=0;

    binaries_up[branch_variable[origin_worker]]=1;

    
    child_down.initialize(binaries_down, num_bins, parent_value);
 
    child_up.initialize(binaries_up, num_bins, parent_value);
    
    
    delete [] binaries_down;
    delete [] binaries_up;
    
    // Create nodestruct for the up-child
    nodestruct* childstruct_up = new nodestruct[1];
    
    // Set previous node of the up-child to be the down-child and the next to be the next of the branched node
    childstruct_up->node = child_up;
    
    cout<<" Nodes created "<<childstruct_up->node.get_tag()<<endl;
    /*
    nodestruct *list_node = new nodestruct[1];
    list_node = node_to_branch->next;
    
    
    int count_list = 0;

    // If the branch node is the last node in the list
    if (list_node == NULL) {
        childstruct_up->next = NULL;
        childstruct_up->previous = node_to_branch;
        
        // Replace branched node with the down-child and set next node to be the up-child
        node_to_branch->node = child_down;
        node_to_branch->next = childstruct_up;
    
    }
    
    
    // If there are nodes after the branch node
    else {
        while (list_node != NULL) {
            count_list++;
        // If the parent value is larger than the next nodes bound they should switch places
            if (parent_value > list_node->node.get_lowerbound()) {
                // If the branch node is the first in the list, set the next as the first_node
                if (node_to_branch->previous == NULL) {
                    first_node = list_node;
                    list_node->previous = NULL;
                }
                // Else, set the next in front of the branched node
                else {
                   node_to_branch->previous->next = list_node;
                  list_node->previous = node_to_branch->previous;
                }
                
                node_to_branch->previous = list_node;
                
                
                if (list_node->next == NULL){
                    node_to_branch->next = NULL;
                    list_node->next = node_to_branch;
                    list_node = NULL; 
                }
                else{
                    node_to_branch->next = list_node->next;
                    list_node->next->previous = node_to_branch;
                    list_node->next = node_to_branch;
                    list_node = node_to_branch->next;
                }
                
                // If we get to the end of the list, branch the node
                if (list_node == NULL) {
                    childstruct_up->next = NULL;
                    childstruct_up->previous = node_to_branch;

                    // Replace branched node with the down-child and set next node to be the up-child
                    node_to_branch->node = child_down;
                    node_to_branch->next = childstruct_up;
  

                    break;
                }
            }
            // If the parent value is lower than the next node's bound, place it in front
            else {
               childstruct_up->next = list_node;
               childstruct_up->previous = node_to_branch;
          node_to_branch->next->previous =  childstruct_up;

               // Replace branched node with the down-child and set next node to be the up-child
                 node_to_branch->node = child_down;
                 node_to_branch->next = childstruct_up;
                 break;
            }
        } 
    }
    */
    
    if (node_to_branch->next != NULL) {
        childstruct_up->next = node_to_branch->next;
        node_to_branch->next->previous =  childstruct_up;

    }
    else {
        childstruct_up->next = NULL;
    }
    
    childstruct_up->previous = node_to_branch;
 
        
    // Replace branched node with the down-child and set next node to be the up-child
    node_to_branch->node = child_down;
    node_to_branch->next = childstruct_up;

    nodes_open++;
    nodes_open++;
}

// Possible memory problem if structs are not deleted?
void bbclass::prune_node(nodestruct *node_to_prune) {
    //There is just one in the list
    if (node_to_prune->previous == NULL and node_to_prune->next == NULL) {
    }
    // If the node to prune is first in the list, set the first_node to be the next node
    else if (node_to_prune->previous == NULL ) {
        first_node = node_to_prune->next;
        node_to_prune->next = NULL;
        first_node->previous = NULL;
    }
    // If the node is last of the list, set the previous node's next to NULL
    else if (node_to_prune->next == NULL) {

        node_to_prune->previous->next = NULL;
        node_to_prune->previous = NULL;
    }
    // Else, the node to prune is in the middle of the list, remove it
    else {
   
    // Set parent node's next node to the pruned node's next node
    node_to_prune->previous->next = node_to_prune->next;
     
    // Set next node of pruned node to null
     
    node_to_prune->next->previous = node_to_prune->previous;

    //delete[] node_to_prune;
    node_to_prune->previous = NULL;
    node_to_prune->next= NULL;
    
    if (incumbent_node != node_to_prune) {
        delete[] node_to_prune;
    }
    
    }
  
}

int bbclass::show_nodes(){
nodestruct *pointer = first_node;
int *fixed = new int[num_bins];
int counter=0;
cout<<"**********************************"<<endl;
cout<<"0: These are my nodes: "<<endl;
while(pointer != NULL && counter < nodes_open){
    pointer->node.get_binaries(fixed,num_bins);
cout<<"listed node "<<counter++<<

", lower bound: "<< pointer->node.get_lowerbound()<<
", value: "<< pointer->node.get_value()<<
", tag: "<< pointer->node.get_tag()<<endl;

/*for(int i=0;i<num_bins;i++) {
    if (fixed[i] == 1) {
    cout<<"I have variable "<<i<<" fixed to 1!!! "<<endl;
    }
    else if (fixed[i] == 0) {
    cout<<"I have variable "<<i<<" fixed to 0!!! "<<endl;
 }
}
 */




pointer = pointer->next;
}

delete [] fixed;
delete[] pointer;
cout<<"**********************************"<<endl;
return counter;
}

/*
 
 
 */