/* 
 * File:   bbclass.hpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 *
 * Created on April 3, 2014, 8:26 PM
 */

#ifndef BBCLASS_HPP
#define	BBCLASS_HPP

#include "nodeclass.hpp"

#include "mpi.h"
#include "xprs.h"

#include <stddef.h>
#include <cstdio>


struct nodestruct {   //this struct stores one node, and pointers to the next and previous node in the list
    nodeclass node;
    nodestruct *next;
    nodestruct *previous;
};


class bbclass {
public:
    bbclass();
    bbclass(const bbclass& orig);
    virtual ~bbclass();
    
    void initialize(char *, MPI::Intracomm); // initialize the bbtree with all binary variables relaxed
    int optimize();     //the optimize function called from main

    void branch_node(nodestruct *, int);        //this will take one node and branch it, creating two child nodes
    void prune_node(nodestruct *);  //prune if a node's bound/solution is infeasible or worse than the incumbent.
    void send_notification(int, int);   //send a notification to a worker, 1 means solve, 0 means terminate;
    void send_binaries(int *, int);   //send the status of the binaries of a node to a worker to solve;
    void receive_solution(int); // receives the solution

    int show_nodes (); //print list of open nodes
    
private:
    MPI::Intracomm communicator;
    MPI::Request request;
    MPI::Status status;
    
    nodestruct *first_node;   //the beginning of the linked list
    int numProc; // number of processors available
    int myRank; //rank of a processor
    int nodes_open;     //number of nodes in the list 
    nodestruct *incumbent_node;   //the incumbent node, if any, found so far.
    int num_bins; // Number of binary variables in the problem
    int *solution_status; // Status of the solution from each worker
    double *solution_value; // Value of the solution from each worker
    int *branch_variable; // Branch variable from each worker
    double lower_bound; // current lower bound of the BB tree
    double initial_time; // Time when algorithm starts
};

#endif	/* BBCLASS_HPP */

