/* 
 * File:   nodeclass.hpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 *
 * Created on April 8, 2014, 9:08 PM
 */

#ifndef NODECLASS_HPP
#define	NODECLASS_HPP

#include <float.h>

#include <stddef.h>
#include <cstdio>

class nodeclass {
public:
    nodeclass();
    nodeclass(const nodeclass& orig);
    virtual ~nodeclass();
    
    void initialize(int *, int, double);   //custom constructor
    char get_status(){return status;};
    double get_value(){return value;};
    int *get_binaries(){return fixed_binaries;};
    void get_binaries(int *, int);     
    int get_processor(){return processor;};
    double get_lowerbound(){return lower_bound;};
    int get_tag(){return node_tag;};
    
    void set_status(char p_status){status = p_status;};
    void set_value(double p_value){value = p_value;};
    void set_processor(int rank){processor = rank;};
    void set_lowerbound(double lbound){lower_bound = lbound;};

    static int master_tag;
    
private:
    int node_tag;               // id of node
    
    int *fixed_binaries;     //an array of values to indicate whether each binary has been fixed to 1, 0, or relaxed 
    char status;             //the status of the node, solved or unsolved
    
    double value;           //the value of the solution, if solved and feasible
        
    int num_bins;       // number of binary variables in the problem
    int processor;      // the node should know which processor, if any, is handling it
    double lower_bound; // Value of parent node, optimistic bound on this node

};

#endif	/* NODECLASS_HPP */

