/* 
 * File:   workerclass.hpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 *
 * Created on April 3, 2014, 8:27 PM
 */

#ifndef WORKERCLASS_HPP
#define	WORKERCLASS_HPP

#include "xprs.h"
#include "mpi.h"

#include <cstdlib>

class workerclass {
public:
    workerclass();
    workerclass(const workerclass& orig);
    virtual ~workerclass();
    
    void initialize(char *, MPI::Intracomm);  //initializes the problem reading the file
    int optimize();   //calls xpress to optimize the problem, function run from main
    void receive_notification();   //receives a message from the master, either a solve request (1) or a finalization (0)
    void receive_binaries();   //receives status of binary variables from the master
    void return_solution();   //returns the worker solution to the master
    void finalize();    //if there's anything to finalize
    
private:
    
    XPRSprob prob;    //this will store the Xpress object which will be solved
    
    MPI::Intracomm communicator;
    MPI::Request request;
    MPI::Status status;
    
    int rank;           // rank of the worker
    int nCols;          // number of variables in the problem
    int num_bins;       //we need to know how may binary variables are there 
    int *binaries;      //the status of the binary variables in the problem
    int *indexBinVar;  //the variable index of each binary variable
   
    int instruction; // instruction from the master, 0 for terminate, 1 for solve
    int solution_status; //whether the solution is infeasible, feasible or integer feasible
    double solution_value; //the value of the feasible or integer feasible solution
    double *x; //the value of the variables in the solution
    int branch_variable; // which binary variable to branch on
};

#endif	/* WORKERCLASS_HPP */

