/* 
 * File:   bbclass.hpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 *
 * Created on April 3, 2014, 8:26 PM
 */

#ifndef BBCLASS_HPP
#define	BBCLASS_HPP

#include "mpi.h"
#include "xprs.h"

#include <stddef.h>
#include <cstdio>
#include <iostream>

#include <vector>

using namespace std;

class bbclass {
public:
    bbclass();
    bbclass(const bbclass& orig);
    virtual ~bbclass();
    
    void initialize(char *, MPI::Intracomm);  //initializes the problem reading the file and other stuff
    
    int optimize();       //the optimize function called from main

    void send_notification(int, int);   //send a notification to a worker, 1 means solve, 0 means terminate;
    void send_binaries(int *, int);   //send the status of the binaries of a node to a worker to solve;
    void receive_solution(int); // receives the solution

private:
    MPI::Intracomm communicator;
    MPI::Request request;
    MPI::Status status;
    
    int numProc; // number of processors available
    int myRank; //rank of a processor
    int incumbent_worker; //the worker who returned the current incumbent solution
    double incumbent_value;   //the incumbent node, if any, found so far.
    int num_bins; // Number of binary variables in the problem
    int *solution_status; // Status of the solution from each worker
    double *solution_value; // Value of the solution from each worker
};

#endif	/* BBCLASS_HPP */

