/* 
 * File:   workerclass.cpp
 * Author: Kristine Ivarsøy and Ida Elise Solhaug
 * 
 * Created on April 3, 2014, 8:27 PM
 */

#include "workerclass.hpp"

#include <cmath>
#include <sstream>


using namespace std;

workerclass::workerclass() {
    
}

workerclass::workerclass(const workerclass& orig) {

}

workerclass::~workerclass() {
}

void workerclass::initialize(char* argv, MPI::Intracomm comm) {
    
    communicator=comm;
    
    rank = communicator.Get_rank();
    
    //allocate memory using the dedicated function
     XPRScreateprob(&prob);
    //read the problem from the file path provided since main
    XPRSreadprob(prob,argv,"");
    
     // Create logfile
    string logfile;
    stringstream ss;
    ss << rank << "log_60_AF_TM.txt";
    logfile = ss.str();
    XPRSsetlogfile(prob, logfile.c_str());

    //Get number of variables in the problem
    XPRSgetintattrib(prob, XPRS_COLS, &nCols);
   
    char *types = new char[nCols];
    int *temp_bin_indices = new int[nCols];
    
    //Store type of each variable in the problem
    XPRSgetcoltype(prob, types, 0, nCols - 1);
    
    num_bins=0;
    // Count number of binary variables in the problem
    for (int i = 0; i < nCols; i++) {
        if (types[i] == 'B') {
            temp_bin_indices[num_bins] = i;
            num_bins++;
        }
    }
    
    //copy the indices into the proper field and destroy the temporal array
    indexBinVar = new int[num_bins];
    for(int i=0;i<num_bins;i++){
        indexBinVar[i] = temp_bin_indices[i];
    }
    
    delete[] temp_bin_indices;
    delete[] types;
}

int workerclass::optimize(){
    //receive from the master a notification of what to do: 1 to solve, 0 to terminate
    receive_notification();

    //if I am told to solve
    while (instruction != 0){

        //receive from the master a set of fixed variables
        binaries = new int[num_bins];

        receive_binaries();

        //change the upper bounds according to the fixing vector
        double *bounds = new double[2*num_bins];
        char *type_of_bound = new char[2*num_bins];
        int *var_index = new int[2*num_bins];
        
        int nn = 0;
        //fill the array to change the upper bounds:
        for(int i=0;i<num_bins;i++){
            if (binaries[i] == -1) {
                type_of_bound[nn]='U';
                bounds[nn] = 1;
                var_index[nn] = indexBinVar[i];
                nn++;
                type_of_bound[nn]='L';
                bounds[nn] = 0;
                var_index[nn] = indexBinVar[i];
                nn++;
            }
            else if (binaries[i] == 1) {
                type_of_bound[nn]='B';
                bounds[nn] = 1;
                var_index[nn] = indexBinVar[i];
                nn++;

            }
            
            else if (binaries[i] == 0) {
                type_of_bound[nn]='B';
                bounds[nn] = 0;
                var_index[nn] = indexBinVar[i];
                nn++;
            }
            
        }
        
        XPRSchgbounds(prob, nn, var_index, type_of_bound, bounds);

    
        delete[] type_of_bound;
        delete[] bounds;
        delete[] var_index;
 
      XPRSsetintcontrol(prob, XPRS_MAXTIME, -2250;
        
        cout<<rank<<": Going to start optimize now! "<<endl;

 
        //Solve the integer problem with the fixed variables
        XPRSmipoptimize(prob, "");
        
        cout<<rank<<": Done optimizing! "<<endl;

        
         // Create filename for solution file
        string solfile;
        stringstream rr;
        rr << rank << "_sol_60_AF_TM" << MPI::Wtime();
        solfile = rr.str();
      
        //Write solution to file
        XPRSwriteprtsol(prob, solfile.c_str(), "");

       cout<<rank<<": Solution written down "<<endl;

        
         // Get status of the solution
        XPRSgetintattrib(prob,XPRS_MIPSTATUS,&solution_status);
       
        if (solution_status == XPRS_MIP_INFEAS) {
            solution_status = 0;
            cout<<rank<<": MIP is infeasible!"<<endl;
        }
        else if (solution_status == XPRS_MIP_SOLUTION) {
            solution_status = 1;
            XPRSgetdblattrib(prob, XPRS_MIPOBJVAL, &solution_value);
            cout<<rank<<": Integer solution found, but search is not complete. Solution is "<<solution_value<<endl;
        }
        
        else if (solution_status == XPRS_MIP_OPTIMAL) {
            solution_status = 1;
            XPRSgetdblattrib(prob, XPRS_MIPOBJVAL, &solution_value);
            cout<<rank<<": MIP is optimal with solution "<<solution_value<<endl;
        }
        
        else if (solution_status == XPRS_MIP_NO_SOL_FOUND ) {
            solution_status = 0;
            cout<<rank<<": Search is complete, but no solution found"<<endl;
        }
        
        else if (solution_status == XPRS_MIP_LP_NOT_OPTIMAL ) {
            solution_status = 0;
            cout<<rank<<": LP has not been optimized."<<endl;
        }
        else if (solution_status == XPRS_MIP_NOT_LOADED ) {
            solution_status = 0;
            cout<<rank<<": Problem was not loaded."<<endl;
        }
        else if (solution_status == XPRS_MIP_LP_OPTIMAL ) {
            solution_status = 0;
            XPRSgetdblattrib(prob, XPRS_LPOBJVAL, &solution_value);
            cout<<rank<<": Optimization halted after LP optimization. LP bound was "<<solution_value<<endl;
        }
       
        else {
            solution_status = 0;
            cout<<"Something strange happened..."<<endl;
        }
        //Return the solution to the master
        return_solution();
        
        
        cout<<rank<<": solution returned : "<<solution_status<<endl;

        delete[] binaries;  
        delete[] indexBinVar;
        
        XPRSdestroyprob(prob);
        
        // Receive new instruction
        instruction=0;
        
    }
    //else I am told to terminate
    finalize();
    
    return 0;
    
}

void workerclass::receive_notification(){
    communicator.Recv(&instruction, 1, MPI::INT, 0, 0, status);
}

void workerclass:: receive_binaries() {
    communicator.Recv(binaries, num_bins, MPI::INT, 0, 1, status);
}

void workerclass::return_solution(){
    request=communicator.Issend(&solution_status, 1, MPI::INT, 0, 2);
    request.Wait(status);
    // If an integer solution was found, send the value
    if (solution_status == 1) {
        request=communicator.Issend(&solution_value, 1, MPI::DOUBLE, 0, 3);
        request.Wait(status);
    }
}

void workerclass::finalize(){
        // delete arrays
}